// file name : file_io.cpp
//
// file i/o routines for FreePCB
//
#include "stdafx.h" 

// parse a string into fields, where fields are delimited by whitespace
// if a field is enclosed by "", embedded spaces are allowed
//
/*=============================================================================

=============================================================================*/

int ParseStringFields( CString * str, CArray<CString> * field )
{
	field->SetSize(0);
	enum STATE { NONE, INFIELD, INQUOTE };
	STATE state = NONE;
	int nfields = 0;
	CString field_str;
	LPCTSTR		s=*str;
	int			len=lstrlen(s);

	for( int i=0; i<=len; i++ )
	{
		TCHAR	c;

		if( i == len )
			c = _T('\n');								// force eol
		else
			c = s[i];
		if( c == _T(' ') || c == _T('\t') || c == _T('\n'))
		{
			// whitespace char
			if( state == INQUOTE )
				field_str.AppendChar( c );
			else if( state == INFIELD )
			{
				// end field
				field->SetSize( nfields+1 );
				(*field)[nfields] = field_str;
				field_str = "";
				nfields++;
				state = NONE;
			}
		}
		else if( state == INQUOTE && c == _T('\"'))
		{
			// end quoted field
			field->SetSize( nfields+1 );
			(*field)[nfields] = field_str;
			field_str = "";
			nfields++;
			state = NONE;
		}
		else if( state == INQUOTE || state == INFIELD )
		{
			// non-whitespace char in field
			field_str.AppendChar( c );
		}
		else
		{
			// non-whitespace char, not in field
			if( c == _T('\"'))
				state = INQUOTE;	// start next field in quotes
			else
			{
				state = INFIELD;	// start next field (no quotes)
				field_str.AppendChar( c );
			}
		}
	}
	return nfields;
}

/*=============================================================================

=============================================================================*/

// parse string in format "keyword: param1 param2 param3 ..."
//
// if successful, returns number of params+1 and sets key_str to keyword,
// fills array param_str[] with params
//
// ignores whitespace and trailing comments beginning with "/"
// if the entire line is a comment or blank, returns 0
// if param is a string, must be enclosed in " " if it includes spaces and must be first param
//		then quotes are stripped
// if unable to parse, returns -1
//

int ParseKeyString( CString * str, CString * key_str, CArray<CString> * param_str )
{
	int pos = 0;
	int last_pos;
	int np = 0;
	for( int ip=0; ip<param_str->GetSize(); ip++ )
		(*param_str)[ip] = "";

	str->Trim();
	char c = str->GetAt(0);
	if( str->GetLength() == 0 || c == _T('/'))
		return 0;
	CString keyword = str->Tokenize(_T(" :,\n\t"), pos );
	last_pos = pos;
	if( keyword.IsEmpty())
		return -1;
	// keyword found
	keyword.Trim();
	*key_str = keyword;
	// now get params
	if( keyword.GetLength() == str->GetLength() )
		return 1;
	CString param;
	CString right_str = str->Right( str->GetLength() - keyword.GetLength() - 1);
	right_str.Trim();
	pos = 0;
	BOOL bQuotedStr = TRUE;
	if( right_str.Left(2) == _T("\"\""))
	{
		// quoted string is blank
		right_str = right_str.Right( right_str.GetLength()-2 );
	}
	else if( right_str[0] == _T('\"'))
	{
		// param starts with ", remove it and tokenize to following "
		param = right_str.Tokenize(_T("\""), pos );
	}
	else
	{
		param = right_str.Tokenize(_T(" ,\n\t"), pos );
		bQuotedStr = FALSE;
	}
	while( param != _T("") || bQuotedStr )
	{
		bQuotedStr = FALSE;
		param.Trim();
		param_str->SetAtGrow( np, param );
		// now test for next parameter starting with "
		last_pos = pos;
		param = right_str.Tokenize( _T(" ,\n\t"), pos );
		param.Trim();
		if( param[0] == _T('\"'))
		{
			// retokenize to get rid of ""
			pos = last_pos;
			param = right_str.Tokenize( _T("\""), pos );
			param = right_str.Tokenize( _T("\""), pos );
		}
		np++;
	}

	return np+1;
}

/*=============================================================================

=============================================================================*/
// my_atoi ... calls atoi and throws * CString exception if error
// if input string ends in "MIL", "mil", "MM", or "mm" convert to NM
//
int my_atoi(LPCTSTR lpszString)
{
	CString str = lpszString;
	BOOL mil_units = FALSE;
	BOOL mm_units = FALSE;
	int len = str.GetLength();
	if( len > 3 && ( str.Right(3) ==_T("MIL") || str.Right(3) == _T("mil")) )
	{
		mil_units = TRUE;
		str = str.Left(len-3);
		len = len-3;
	}
	else if( len > 2 && ( str.Right(2)==_T("MM") || str.Right(2)==_T("mm")) )
	{
		mm_units = TRUE;
		str = str.Left(len-2);
		len = len-2;
	}
	int test = _ttoi( str );
	if( test == 0 )
	{
		for( int i=0; i<len; i++ )
		{
			TCHAR	c = str[i];
			if( !( c == _T('+') || c == _T('-') || ( c >= _T('0')) ) )
			{
				CString * err_str = new CString;
				err_str->Format(_T("my_atoi() unable to convert string \"%s\" to int"), lpszString);
				throw err_str;
			}
		}
	}
	if( mil_units )
		test = test * NM_PER_MIL;
	else if( mm_units )
		test = test * 1000000;
	return test;
}

/*=============================================================================

=============================================================================*/
// my_atof ... calls atof and throws * CString exception if error
// if input string ends in "MIL", "mil", "MM", or "mm" convert to NM
//
double my_atof(LPCTSTR lpszString)
{
	CString str = lpszString;
	BOOL mil_units = FALSE;
	BOOL mm_units = FALSE;
	int len = str.GetLength();
	if( len > 3 && ( str.Right(3) == _T("MIL") || str.Right(3) == _T("mil")) )
	{
		mil_units = TRUE;
		str = str.Left(len-3);
		len = len-3;
	}
	else if( len > 2 && ( str.Right(2) == _T("MM") || str.Right(2) == _T("mm")) )
	{
		mm_units = TRUE;
		str = str.Left(len-2);
		len = len-2;
	}
	double test = _tstof( str );
	if( test == 0.0 )
	{
		for( int i=0; i<len; i++ )
		{
			TCHAR	c = str[i];

			if( !( c == _T('.') || c == _T('+') || c == _T('-') || c == _T('0') || c == _T(' ')) )
			{
				CString * err_str = new CString;
				err_str->Format(_T("my_atof() unable to convert string \"%s\" to double"), lpszString);
				throw err_str;
			}
		}
	}
	if( mil_units )
		test = test * NM_PER_MIL;
	else if( mm_units )
		test = test * 1000000.0;
	return test;
}
